# @rickosborne/foundation

Basic data structures and algorithms.
Built on [@rickosborne/typical](https://www.npmjs.com/package/@rickosborne/typical) for types, and [@rickosborne/guard](https://www.npmjs.com/package/@rickosborne/guard) for guards.

***

## API

### Classes

#### ArrayBinaryMinHeap

<a id="api-arraybinaryminheap"></a>

```typescript
 class ArrayBinaryMinHeap<T> implements Queue<T> 
```

Simple implementation of a binary min-heap, meaning low values are sorted before high values, using the supplied comparator. Invert your comparator if you want a max-heap.


#### PriorityArray

<a id="api-priorityarray"></a>

```typescript
 class PriorityArray<T> implements Queue<T> 
```

An implementation of a priority queue backed by a JS array, using the given comparator. Sorts low-to-high, so you will need to invert comparator if you want high-to-low.


#### PriorityLinkedList

<a id="api-prioritylinkedlist"></a>

```typescript
 class PriorityLinkedList<T> implements Queue<T> 
```

Implementation of a priority queue backed by a linked list, using the given comparator. Sorts low-to-high, so you will need to invert comparator if you want high-to-low.


#### SkipList

<a id="api-skiplist"></a>

```typescript
 class SkipList<T> implements Queue<T> 
```

Basic skip list implementation, using the given comparator to keep values in order, low-to-high. If you want high-to-low you will need to invert your comparator.


### Functions

#### addProperties

<a id="api-addproperties"></a>

```typescript
addProperties: <T, K extends string, R extends Record<K, TypedPropertyDescriptor<unknown>>>(target: T, descriptors: R) => Resolve<T & PropertiesFromDescriptors<R>>
```

A more type-friendly version of <kbd>Object.defineProperties</kbd>. See [Object.defineProperties on MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/defineProperties).


#### addProperty

<a id="api-addproperty"></a>

```typescript
addProperty: <T, K extends string, V, D extends TypedPropertyDescriptor<V>>(target: T, key: K, descriptor: D) => Resolve<T & PropertyFromDescriptor<K, D>>
```

A more type-friendly version of <kbd>Object.defineProperty</kbd>. See [Object.defineProperty on MDN](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/defineProperty).


#### arrayBinaryMaxHeap

<a id="api-arraybinarymaxheap"></a>

```typescript
arrayBinaryMaxHeap: <T>(comparator: Comparator<T>) => ArrayBinaryMinHeap<T>
```

Build a max-heap using the given comparator. Will invert the comparator and build a min-heap, which produces the same outcomes.


#### arrayBinaryMinHeap

<a id="api-arraybinaryminheap"></a>

```typescript
arrayBinaryMinHeap: <T>(comparator: Comparator<T>) => ArrayBinaryMinHeap<T>
```

Build a min-heap using the given comparator.


#### arrayEq

<a id="api-arrayeq"></a>

```typescript
arrayEq: <T>(a: T[], b: T[], predicate?: ((a: T, b: T, index: number) => boolean)) => boolean
```

Simple comparison to see if two arrays are equivalent.


#### asFraction

<a id="api-asfraction"></a>

```typescript
asFraction: (num: number, percentError?: number) => FracTuple
```

Try to find a human-readable fraction for the given decimal value, within a given percentage error. See [the original algorithm by Sjaak](https://stackoverflow.com/a/45314258).


#### asIndexKey

<a id="api-asindexkey"></a>

```typescript
asIndexKey: (key: string) => number | undefined
```

Try to convert the string-format key into an array index key.


#### binaryIndexOf

<a id="api-binaryindexof"></a>

```typescript
binaryIndexOf: <T>(value: T, items: T[], comparator: Comparator<T>, config?: BinaryIndexOfConfig) => SearchResult
```

Perform a binary search through the given list for the given value using the given comparator.


#### chainableComparator

<a id="api-chainablecomparator"></a>

```typescript
chainableComparator: <T>(comparator: Comparator<T>) => ChainableComparator<T>
```

Wrap the given comparator so you can chain [ChainableComparator](#api-chainablecomparator) functions from it.


#### closeTo

<a id="api-closeto"></a>

```typescript
closeTo: (a: number, b: number, epsilon?: number) => boolean
```

Test for whether two numbers are close enough to each other, within an acceptable margin.


#### closeToZero

<a id="api-closetozero"></a>

```typescript
closeToZero: (n: number, epsilon?: number) => n is 0
```

Test for whether a number is close enough to zero, within an acceptable margin.


#### comparatorBuilder

<a id="api-comparatorbuilder"></a>

```typescript
comparatorBuilder: <T>() => ComparatorBuilder<T>
```

Start building a comparator.


#### concatRegExp

<a id="api-concatregexp"></a>

```typescript
concatRegExp: (...patterns: RegExp[]) => RegExp
```

Concatenate the patterns (and flags) of regular expressions so they will match all at once. This does not attempt to fix any problems, such as duplicate named capture groups or incorrectly-numbered backreferences.


#### deepCopy

<a id="api-deepcopy"></a>

```typescript
deepCopy: <T>(source: T, options?: DeepCopyOptions) => T
```

Create a deep copy of an object. This is not a particularly robust check, and won't copy things like class instances.


#### deepSort

<a id="api-deepsort"></a>

```typescript
deepSort: <T>(obj: T, options?: DeepSortOptions) => T
```

Sort an object, and any objects nested through plain objects or arrays, by key, so when JSON stringified you will get a deterministic output.


#### doubleLinkedList

<a id="api-doublelinkedlist"></a>

```typescript
doubleLinkedList: <T>() => DoubleLinkedList<T>
```

Build a doubly-linked list.


#### formatBytes

<a id="api-formatbytes"></a>

```typescript
formatBytes: (bytes: number | bigint, base?: 1024 | 1000, precision?: number) => string
```

Pretty-print a byte value with low enough detail to be quickly human-readable.


#### jsonPathAppend

<a id="api-jsonpathappend"></a>

```typescript
jsonPathAppend: (path: string, key: string | number | symbol, options?: JsonPathAppendOptions) => string
```

Append the given property key to the given JSON Path expression. This does not attempt to validate anything about the path or the key, only formats it according to very basic rules. Technically, the spec doesn't handle symbol values, so if you give it a path with a symbol you get an expression which other JSON Path parsers can't handle. However, it does use the correct function syntax, so it should still at least parse correctly.



#### memoize

<a id="api-memoize"></a>

```typescript
memoize: <T>(supplier: Supplier<T>) => Supplier<T>
```

Shorthand for [memoizeSupplier](#api-memoizesupplier).


#### memoizeSupplier

<a id="api-memoizesupplier"></a>

```typescript
memoizeSupplier: <T>(supplier: Supplier<T>) => Supplier<T>
```

Memoize the given supplier so that it is called at most once.


#### optionalRegExp

<a id="api-optionalregexp"></a>

```typescript
optionalRegExp: (pattern: RegExp) => RegExp
```

Modify a regular expression's pattern to make the entire expression optional, keeping any start/end anchors in place


#### parseLocalDate

<a id="api-parselocaldate"></a>

```typescript
parseLocalDate: (text: string, config?: ParseLocalDateConfig) => Date
```

Parse an ISO-8860-like date, but do so assuming that the time is intended to be local, not UTC. But if allowed, the date could override this with specific time zone info.



#### priorityArray

<a id="api-priorityarray"></a>

```typescript
priorityArray: <T>(comparator: Comparator<T>) => Queue<T>
```

Build a priority queue which uses an array as a backing store.


#### priorityLinkedList

<a id="api-prioritylinkedlist"></a>

```typescript
priorityLinkedList: <T>(comparator: Comparator<T>) => Queue<T>
```

Build a priority queue which uses a linked list as a backing store.


#### reQuote

<a id="api-requote"></a>

```typescript
reQuote: (text: string, options?: ReQuoteOptions) => string
```

Given a text, either wrap it in quotes which make the most sense for the text, or change the quotes to a different set.


#### shuffle

<a id="api-shuffle"></a>

```typescript
shuffle: <T>(list: T[], { into, random01 }?: ShuffleConfig<T>) => T[]
```

Fisher-Yates shuffle. Defaults to in-place sorting, but can sort into a new one via <kbd>into</kbd>.



#### singleLinkedList

<a id="api-singlelinkedlist"></a>

```typescript
singleLinkedList: <T>() => List<T>
```

Builder for a single-linked list.


#### skipList

<a id="api-skiplist"></a>

```typescript
skipList: <T>(comparator: Comparator<T>) => SkipList<T>
```

Build a priority queue using a skip list implementation.


#### toShuffled

<a id="api-toshuffled"></a>

```typescript
toShuffled: <T>(list: T[], config?: Omit<ShuffleConfig<T>, "into"> | undefined) => T[]
```

Wrapper for the [shuffle](#api-shuffle) function which shuffles into a new array, leaving the original intact.


#### window2

<a id="api-window2"></a>

```typescript
 function window2<T>(items: Readonly<T[]>): Generator<Readonly<[T, T, number, number]>, void, undefined>;
```

Generator for a sliding window of pairs of values from the given list. It yields each value in the list in a tuple with its successor. The last item wraps around to use the first item as its successor. The indices of the two values are also yielded after the values. For example:

```typescript
for (const [ x, y, xIndex, yIndex ] of window2([ "a", "b", "c" ])) {
    console.log(x, y, xIndex, yIndex);
}
```


This would output: <pre> a, b, 0, 1 b, c, 1, 2 c, a, 2, 0 </pre>


#### window3

<a id="api-window3"></a>

```typescript
 function window3<T>(items: Readonly<T[]>): Generator<Readonly<[T, T, T, number, number, number]>, void, undefined>;
```

Generator for a sliding window of triplets of values from the given list. It yields each value in the list in a tuple with its predecessor and successor. The first and last items wrap around to use the last and first items as their predecessor/successor. The indices of the values are also provided after the values. The middle value (<kbd>y</kbd> in the example below) is the "current" value, with the first value (<kbd>x</kbd>) being its predecessor and the third (<kbd>z</kbd>) its successor. That current value is the one which will be at index 0 for the first iteration of the loop. For example:

```typescript
for (const [ x, y, z, xIndex, yIndex, zIndex ] of window3([ "a", "b", "c" ])) {
    console.log(x, y, z, xIndex, yIndex, zIndex);
}
```


This would output: <pre> c, a, b, 2, 0, 1 a, b, c, 0, 1, 2 b, c, a, 1, 2, 0 </pre>


#### withLazyProperty

<a id="api-withlazyproperty"></a>

```typescript
withLazyProperty: <Target, Property, Name extends string>(target: Target, name: Name, calculator: (t: Target) => Property) => WithProperty<Target, Name, Property>
```

Add a property to the given object which is calculated lazily, and only once.


#### withNullsFirst

<a id="api-withnullsfirst"></a>

```typescript
withNullsFirst: <T>(comparator: Comparator<T>) => Comparator<T | null | undefined>
```

Wrap the given comparator to return null and undefined values before the others.


#### withNullsLast

<a id="api-withnullslast"></a>

```typescript
withNullsLast: <T>(comparator: Comparator<T>) => Comparator<T | null | undefined>
```

Wrap the given comparator to return null and undefined values after the others.


### Interfaces

#### DeepCopyOptions

<a id="api-deepcopyoptions"></a>

```typescript
export interface DeepCopyOptions 
```

#### DeepSortOptions

<a id="api-deepsortoptions"></a>

```typescript
export interface DeepSortOptions 
```

#### JsonPathAppendOptions

<a id="api-jsonpathappendoptions"></a>

```typescript
export interface JsonPathAppendOptions 
```

### TypeAliases

#### BinaryIndexOfConfig

<a id="api-binaryindexofconfig"></a>

```typescript
type BinaryIndexOfConfig = {
    firstNonUnique?: boolean;
    initialLeft?: number;
    initialRight?: number;
    rangeCheck?: boolean;
};
```

Configuration options for [binaryIndexOf](#api-binaryindexof).


#### ChainableComparator

<a id="api-chainablecomparator"></a>

```typescript
type ChainableComparator<T> = Comparator<T> & {
    get desc(): ChainableComparator<T>;
    get nullsFirst(): ChainableComparator<T | undefined | null>;
    get nullsLast(): ChainableComparator<T | undefined | null>;
};
```

Mixin for a comparator with chaining functions to constrain the behavior of the comparator.


#### ComparatorBuilder

<a id="api-comparatorbuilder"></a>

```typescript
type ComparatorBuilder<T> = {
    build(): Comparator<T>;
    get desc(): ComparatorBuilder<T>;
    get nullsFirst(): ComparatorBuilder<T>;
    get nullsLast(): ComparatorBuilder<T>;
    num(accessor: (t: T) => (undefined | number)): ComparatorBuilder<T>;
    str(accessor: (t: T) => (undefined | string)): ComparatorBuilder<T>;
};
```

Build a comparator for the given type, which goes through the specified checks in the order they are defined.


#### DoubleLinkedList

<a id="api-doublelinkedlist"></a>

```typescript
type DoubleLinkedList<T> = List<T> & {
    get tail(): T | undefined;
    reverseValues(): Generator<T, void, undefined>;
    reverseValuesAndIndexes(): Generator<[T, number], void, undefined>;
    reverseWalk(block: (value: T, index: number) => ListWalkDecision<T>): void;
};
```

Mixin for additional methods available when the linked list is double-linked.


#### FracTuple

<a id="api-fractuple"></a>

```typescript
type FracTuple = [numerator: number, denominator: number];
```

A fraction represented as a tuple of numerator and denominator. The numerator is negative if the value is negative. The denominator is always positive, and at least 1. Both values are integers, though the numerator may be Infinity or NaN as the value requires.


#### GetPropertyDescriptor

<a id="api-getpropertydescriptor"></a>

```typescript
type GetPropertyDescriptor<T> = PropertyDescriptorFlags & {
    get(): T;
    set?: never;
    value?: never;
};
```

Partial property descriptor which includes only an accessor (getter) and not a mutator, and is thus read-only.


#### GetSetPropertyDescriptor

<a id="api-getsetpropertydescriptor"></a>

```typescript
type GetSetPropertyDescriptor<T> = PropertyDescriptorFlags & {
    get(): T;
    set(t: T): void;
};
```

Partial property descriptor which includes both an accessor (getter) and mutator (setter), and is thus read-write.


#### IfReadOnlyDescriptor

<a id="api-ifreadonlydescriptor"></a>

```typescript
type IfReadOnlyDescriptor<T, D extends TypedPropertyDescriptor<unknown>> = D extends ReadOnlyPropertyDescriptor | GetPropertyDescriptor<unknown> ? T : never;
```

Filter type <kbd>T</kbd> based on whether the property descriptor <kbd>D</kbd> would produce a read-only property.


#### IfReadWriteDescriptor

<a id="api-ifreadwritedescriptor"></a>

```typescript
type IfReadWriteDescriptor<T, D extends TypedPropertyDescriptor<unknown>> = D extends ReadOnlyPropertyDescriptor | GetPropertyDescriptor<unknown> | SetPropertyDescriptor<unknown> ? never : T;
```

Filter type <kbd>T</kbd> based on whether the property descriptor <kbd>D</kbd> would produce a read-write (or write-only) property.


#### IfWriteOnlyDescriptor

<a id="api-ifwriteonlydescriptor"></a>

```typescript
type IfWriteOnlyDescriptor<T, D extends TypedPropertyDescriptor<unknown>> = D extends SetPropertyDescriptor<unknown> ? T : never;
```

Filter type <kbd>T</kbd> based on whether the property descriptor <kbd>D</kbd> would produce a write-only property. Note: JS has a concept of a write-only property, and the unit tests cover this. However, TS doesn't have a way of expressing it, at least as of v5.8. So for now, it goes into the read-write bucket, even though you can't actually read from it. The alternative would be to set it to `undefined`, to show you can't read from it ... but then you lose the type info for the setter. Yuck. For now ... The types below will, ahem, mangle the types they produce. Just a little. If you produce a setter-only property, it will end up looking like:

```typescript
const withSetterOnly: WithSetterOnly = addProperty({}, "whatever", { set: (value: string) => void(whatever(value)) });
type WithSetterOnly = {
   whatever: string;
   whateverIsWriteOnly?: undefined;
}
```


So you get a reminder without breaking anything too hard.


#### IntGenerator

<a id="api-intgenerator"></a>

```typescript
type IntGenerator = Generator<number, void, undefined> & {
    toArray(): number[];
};
```

Mixin for a generator which also has the ability to bundle its output into an array.


#### IntGeneratorTo

<a id="api-intgeneratorto"></a>

```typescript
type IntGeneratorTo = {
    toExclusive(stopExclusive: number): IntGenerator;
    toInclusive(stopInclusive: number): IntGenerator;
};
```

Partial for the last step of an [IntRange](#api-intrange) builder.


#### IntRange

<a id="api-intrange"></a>

```typescript
type IntRange = {
    from(startInclusive: number): {
        by(skip: number): IntGeneratorTo;
        get down(): IntGeneratorTo;
        get up(): IntGeneratorTo;
    };
};
```

Integer range generators which (hopefully) make it super clear what your boundary, direction, and increment are.


#### List

<a id="api-list"></a>

```typescript
type List<T> = {
    at(index: number): T | undefined;
    get head(): T | undefined;
    insertAt(index: number, value: T): void;
    get isEmpty(): boolean;
    get length(): number;
    pop(): T | undefined;
    push(value: T): void;
    shift(): T | undefined;
    toArray(): T[];
    unshift(value: T): void;
    values(): Generator<T, void, undefined>;
    valuesAndIndexes(): Generator<[T, number], void, undefined>;
    walk(block: (value: T, index: number) => ListWalkDecision<T>): void;
    readonly [Symbol.toStringTag]: string;
};
```

Basic list-like structure.


#### ListWalkDecision

<a id="api-listwalkdecision"></a>

```typescript
type ListWalkDecision<T> = {
    deleteItem?: boolean;
    insertAfter?: T;
    insertBefore?: T;
    keepWalking: boolean;
};
```

Controls how a [List](#api-list)'s <kbd>walk</kbd> operation behaves after each iteration.


#### ParseLocalDateConfig

<a id="api-parselocaldateconfig"></a>

```typescript
type ParseLocalDateConfig = {
    offsetMinutes?: number;
    time?: boolean;
    timeDelimiters?: string[];
    zone?: boolean;
};
```

Configuration for a [parseLocalDate](#api-parselocaldate) call.


#### PropertiesFromDescriptors

<a id="api-propertiesfromdescriptors"></a>

```typescript
type PropertiesFromDescriptors<R extends Record<string, TypedPropertyDescriptor<unknown>>> = {
    readonly [K in ReadOnlyDescriptorsKeys<R>]: R[K] extends TypedPropertyDescriptor<infer V> ? V : never;
} & {
    [K in ReadWriteDescriptorsKeys<R>]: R[K] extends TypedPropertyDescriptor<infer V> ? V : never;
} & {
    [K in WriteOnlyDescriptorsKeys<R>]: R[K] extends TypedPropertyDescriptor<infer V> ? V : never;
} & {
    [K in RenamedWriteOnlyDescriptorsKeys<R>]?: undefined;
};
```

Produce an object with the properties described by the descriptors. See [RenameIfWriteOnlyDescriptor](#api-renameifwriteonlydescriptor) for details on the mangling of write-only properties.


#### PropertyDescriptorFlags

<a id="api-propertydescriptorflags"></a>

```typescript
type PropertyDescriptorFlags = {
    configurable?: boolean;
    enumerable?: boolean;
    writable?: boolean;
};
```

Basic flags for any property descriptor.


#### PropertyFromDescriptor

<a id="api-propertyfromdescriptor"></a>

```typescript
type PropertyFromDescriptor<K extends string, D extends TypedPropertyDescriptor<unknown>> = D extends TypedPropertyDescriptor<infer V> ? ({
    readonly [key in IfReadOnlyDescriptor<K, D>]: V;
} & {
    [key in IfReadWriteDescriptor<K, D>]: V;
} & {
    [key in IfWriteOnlyDescriptor<K, D>]: V;
} & {
    [key in RenameIfWriteOnlyDescriptor<K, D>]?: undefined;
}) : never;
```

Produce a single-property object type for the given property descriptor. Note: will produce an additional mangled pseudo-property for write-only properties. See [RenameIfWriteOnlyDescriptor](#api-renameifwriteonlydescriptor) for details on the mangling.


#### Queue

<a id="api-queue"></a>

```typescript
type Queue<T> = {
    add(value: T): void;
    peek(): T | undefined;
    remove(value: T): void;
    get length(): number;
    take(): T | undefined;
    toArray(): T[];
    values(): Generator<T, void, undefined>;
};
```

Basic interface for a queue.


#### ReadOnlyDescriptorsKeys

<a id="api-readonlydescriptorskeys"></a>

```typescript
type ReadOnlyDescriptorsKeys<R> = R extends Record<infer K, TypedPropertyDescriptor<unknown>> ? {
    [key in K]: IfReadOnlyDescriptor<key, R[key]>;
}[K] : never;
```

Produce the keys for the given property descriptors record which would produce read-only properties.


#### ReadOnlyPropertyDescriptor

<a id="api-readonlypropertydescriptor"></a>

```typescript
type ReadOnlyPropertyDescriptor = {
    writable: false;
};
```

Partial property descriptor which has been marked as non-writable.


#### ReadWriteDescriptorsKeys

<a id="api-readwritedescriptorskeys"></a>

```typescript
type ReadWriteDescriptorsKeys<R> = R extends Record<infer K, TypedPropertyDescriptor<unknown>> ? {
    [key in K]: IfReadWriteDescriptor<key, R[key]>;
}[K] : never;
```

Produce the keys for the given property descriptors record which would produce read-write (or write-only) properties.


#### RenamedWriteOnlyDescriptorsKeys

<a id="api-renamedwriteonlydescriptorskeys"></a>

```typescript
type RenamedWriteOnlyDescriptorsKeys<R> = R extends Record<infer K, TypedPropertyDescriptor<unknown>> ? {
    [key in string & K]: RenameIfWriteOnlyDescriptor<key, R[key]>;
}[string & K] : never;
```

Produce the <strong>mangled</strong> keys for the given property descriptors record which would produce write-only properties. See [RenameIfWriteOnlyDescriptor](#api-renameifwriteonlydescriptor) for details on the mangling.


#### RenameIfWriteOnlyDescriptor

<a id="api-renameifwriteonlydescriptor"></a>

```typescript
type RenameIfWriteOnlyDescriptor<Key extends string, D extends TypedPropertyDescriptor<unknown>> = D extends SetPropertyDescriptor<unknown> ? `${Key}IsWriteOnly` : never;
```

Mangle the given <kbd>Key</kbd> to <kbd>$\{Key\}IsWriteOnly</kbd> if property descriptor <kbd>D</kbd> would produce a write-only property.


#### ReQuoteOptions

<a id="api-requoteoptions"></a>

```typescript
type ReQuoteOptions = {
    auto?: true;
    quotes?: never;
} | {
    auto?: never;
    quotes?: string;
};
```

#### SearchResult

<a id="api-searchresult"></a>

```typescript
type SearchResult = {
    before?: undefined;
    comparisons: number;
    exists: true;
    index: number;
} | {
    before: number;
    comparisons: number;
    exists: false;
    index?: undefined;
};
```

A result of a search through a list. If the value was found, its index is returned. Otherwise, the <kbd>before</kbd> value is the index before which the value could be inserted. For troubleshooting and instrumentation, also returns the number of comparisons it took to find the value.


#### SetPropertyDescriptor

<a id="api-setpropertydescriptor"></a>

```typescript
type SetPropertyDescriptor<T> = PropertyDescriptorFlags & {
    get?: never;
    set(t: T): void;
};
```

Partial property descriptor which includes only a mutator (setter) and not an accessor (getter), and is thus write-only. Note: TypeScript support for these kinds of properties is very poor as of v5.7.


#### ShuffleConfig

<a id="api-shuffleconfig"></a>

```typescript
type ShuffleConfig<T> = {
    into?: T[] | undefined;
    random01?: () => number;
};
```

Configuration options for the [shuffle](#api-shuffle) function.


#### TypedPropertyDescriptor_2

<a id="api-typedpropertydescriptor-2"></a>

```typescript
type TypedPropertyDescriptor<T> = ValuePropertyDescriptor<T> | GetPropertyDescriptor<T> | SetPropertyDescriptor<T> | GetSetPropertyDescriptor<T>;
```

Union for reasonable configurations of property descriptors which could affect the type for the property.


#### ValuePropertyDescriptor

<a id="api-valuepropertydescriptor"></a>

```typescript
type ValuePropertyDescriptor<T> = PropertyDescriptorFlags & {
    get?: never;
    set?: never;
    value: T;
};
```

Partial property descriptor which contains a static value, instead of an accessor and/or mutator.


#### WriteOnlyDescriptorsKeys

<a id="api-writeonlydescriptorskeys"></a>

```typescript
type WriteOnlyDescriptorsKeys<R> = R extends Record<infer K, TypedPropertyDescriptor<unknown>> ? {
    [key in string & K]: IfWriteOnlyDescriptor<key, R[key]>;
}[string & K] : never;
```

Produce the non-mangled keys for the given property descriptors record which would produce write-only properties.


### Variables

#### DQ

<a id="api-dq"></a>

```typescript
DQ: "\""
```

#### ecmaKeySort

<a id="api-ecmakeysort"></a>

```typescript
ecmaKeySort: Comparator<string>
```

Comparator for Object keys in ECMA "correct" order. Don't read the history on this. It will only depress you.



#### INDEX_KEY_PATTERN

<a id="api-index-key-pattern"></a>

```typescript
INDEX_KEY_PATTERN: RegExp
```

Pattern for an array index key, basically any integer.


#### INFINITY_SIGIL

<a id="api-infinity-sigil"></a>

```typescript
INFINITY_SIGIL = "∞"
```

Text used for displaying human-readable infinity.


#### intRange

<a id="api-intrange"></a>

```typescript
intRange: IntRange
```

Start generating a range of integers.


#### JSON_PATH_ROOT

<a id="api-json-path-root"></a>

```typescript
JSON_PATH_ROOT: "$"
```

#### METRIC_PREFIXES

<a id="api-metric-prefixes"></a>

```typescript
METRIC_PREFIXES: readonly string[]
```

List of metric prefixes, where the index is its base-1000 or base-1024 offset.


#### numberAsc

<a id="api-numberasc"></a>

```typescript
numberAsc: Comparator<number>
```

Numeric comparator which sorts lower values before higher ones.


#### PI_SIGIL

<a id="api-pi-sigil"></a>

```typescript
PI_SIGIL = "π"
```

Text used for displaying human-readable pi.


#### SQ

<a id="api-sq"></a>

```typescript
SQ: "'"
```

#### TIME_DELIMITERS

<a id="api-time-delimiters"></a>

```typescript
TIME_DELIMITERS: string[]
```

Possible delimiters between a date and a time value.


